/**
 * description
 *
 * @author eson
 *2022年7月12日-13:36:24
 */
package com.yuandian.dataflow.statemachine;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.nio.ByteBuffer;
import java.time.Duration;
import java.time.Instant;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.Executor;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.function.Consumer;

import org.reflections.ReflectionUtils;
import org.reflections.Reflections;

import com.alipay.remoting.NamedThreadFactory;
import com.alipay.remoting.exception.CodecException;
import com.alipay.remoting.serialization.SerializerManager;
import com.alipay.sofa.jraft.JRaftUtils;
import com.alipay.sofa.jraft.Node;
import com.alipay.sofa.jraft.RaftGroupService;
import com.alipay.sofa.jraft.Status;
import com.alipay.sofa.jraft.closure.ReadIndexClosure;
import com.alipay.sofa.jraft.conf.Configuration;
import com.alipay.sofa.jraft.entity.PeerId;
import com.alipay.sofa.jraft.entity.Task;
import com.alipay.sofa.jraft.error.RaftError;
import com.alipay.sofa.jraft.error.RemotingException;
import com.alipay.sofa.jraft.option.CliOptions;
import com.alipay.sofa.jraft.option.NodeOptions;
import com.alipay.sofa.jraft.rpc.InvokeCallback;
import com.alipay.sofa.jraft.rpc.InvokeContext;
import com.alipay.sofa.jraft.rpc.RpcClient;
import com.alipay.sofa.jraft.rpc.RpcProcessor;
import com.alipay.sofa.jraft.rpc.impl.BoltRaftRpcFactory;
import com.alipay.sofa.jraft.util.BytesUtil;
import com.alipay.sofa.jraft.util.Endpoint;
import com.alipay.sofa.jraft.util.ThreadPoolUtil;
import com.yuandian.dataflow.statemachine.annotations.MasterRegister;
import com.yuandian.dataflow.statemachine.annotations.WorkerRegister;
import com.yuandian.dataflow.statemachine.closure.GenericClosure;
import com.yuandian.dataflow.statemachine.master.MasterExecute;
import com.yuandian.dataflow.statemachine.operate.Operate;
import com.yuandian.dataflow.statemachine.rpc.OperateProcessor.OperateRequest;
import com.yuandian.dataflow.statemachine.rpc.RaftResponse;
import com.yuandian.dataflow.statemachine.state.State;
import com.yuandian.dataflow.utils.Utils;

import lombok.Getter;
import lombok.Setter;
import lombok.extern.slf4j.Slf4j;

/**
 * description
 *
 * @author eson
 *         2022年7月12日-13:36:24
 */
@Slf4j
public class StateFactory {

    private static StateServer ss;

    public static void startStateServer(String peerstr, Configuration conf) throws Exception {
        if (ss != null) {
            throw new Exception("重复初始化 InitStateServer");
        }
        ss = new StateFactory.StateServer(peerstr, conf);
    }

    public static boolean isLeader() {
        return ss.node.isLeader();
    }

    public static PeerId getLeaderId() {
        return ss.node.getLeaderId();
    }

    public static PeerId getServerId() {
        return ss.cluster.getServerId();
    }

    public static Node getNode() {
        return ss.node;
    }

    public static Node getRaftNode() {
        return ss.cluster.getRaftNode();
    }

    public static RpcClient getRpcClient() {
        return ss.getRpcClient();
    }

    public static RaftGroupService getCluster() {
        return ss.getCluster();
    }

    // 获取状态服务的对象
    public static StateServer getStateServer() {
        return ss;
    }

    public static void readIndexState(GenericClosure<State> closure) {
        ss.readIndexState(closure);
    }

    public static void applyOperate(Operate op, GenericClosure<Operate> closure) {
        ss.applyOperate(op, closure);
    }

    public static void rpcClientInvokeAsync(final Endpoint endpoint, final Object request,
            final InvokeCallback callback, final long timeoutMs)
            throws InterruptedException, RemotingException {
        ss.getRpcClient().invokeAsync(endpoint, request, callback, timeoutMs);
    }

    public static Object rpcClientInvokeSync(final Endpoint endpoint, final Object request, final long timeoutMs)
            throws InterruptedException, RemotingException {
        return ss.getRpcClient().invokeSync(endpoint, request, timeoutMs);
    }

    public static Object rpcClientInvokeSync(final Endpoint endpoint, final Object request, final InvokeContext ctx,
            final long timeoutMs) throws InterruptedException, RemotingException {
        return ss.getRpcClient().invokeSync(endpoint, request, ctx, timeoutMs);
    }

    @Getter
    @Setter
    public static class StateServer {

        RpcClient rpcClient;

        private Node node;
        private RaftGroupService cluster;
        private StateMachine fsm;

        private String groupId = "dataflow";
        private Executor readIndexExecutor = createReadIndexExecutor();

        public StateServer(String addr, Configuration conf) {
            // String[] peers = new
            // String[]{"localhost:4440","localhost:4441","localhost:4442"};
            // String[] sprPeers = new String[]{"3440","3441","3442"};

            // var peeridstr = peers[Integer.parseInt(serverId)];
            // var sprPort = sprPeers[Integer.parseInt(args[0])];

            // String groupId = "jraft";

            // conf =
            // JRaftUtils.getConfiguration("localhost:4440,localhost:4441,localhost:4442");



            PeerId serverId = JRaftUtils.getPeerId(addr);
            int port = serverId.getPort();

            NodeOptions nodeOptions = new NodeOptions();

            nodeOptions.setElectionTimeoutMs(1000);
            nodeOptions.setSnapshotLogIndexMargin(3600);
            nodeOptions.setInitialConf(conf);

            File RaftDataFile = new File(String.format("./raftdata/%d", port));
            log.info("mkdirs: {}", RaftDataFile.mkdirs());

            nodeOptions.setLogUri(String.format("./raftdata/%d/logs", port));
            nodeOptions.setRaftMetaUri(String.format("./raftdata/%d/rafts", port));
            nodeOptions.setSnapshotUri(String.format("./raftdata/%d/snaps", port));
            fsm = new StateMachine(); // 状态实例初始化
            nodeOptions.setFsm(fsm);

            cluster = new RaftGroupService(groupId, serverId, nodeOptions);

            // 扫描注解RaftProccessor 注册
            var now = Instant.now();
            HashMap<String, Class<?>> scansMap = new HashMap<>();
            var traces = Thread.currentThread().getStackTrace();
            var clsName = traces[traces.length - 1].getClassName();
            var packName = clsName.substring(0, Utils.indexOfAsPossible(clsName, ".", 3));
            log.info("获取 {} -> {} 下包的所有注解",clsName, packName );

            var refl = new Reflections(packName);
            Set<Class<?>> scans = refl.getTypesAnnotatedWith(WorkerRegister.class);
            

            scans.forEach((pRaftClass) -> {
                scansMap.put(pRaftClass.getName(), pRaftClass);
            });
            log.info("扫描注解的时间(scan annotations cost time): {} ms", Duration.between(now, Instant.now()).toMillis()) ;
            scansMap.forEach((name, pRaftClass) -> {
                try {
                    cluster.getRpcServer()
                            .registerProcessor((RpcProcessor<?>) pRaftClass.getDeclaredConstructor().newInstance());
                } catch (InstantiationException | IllegalAccessException | IllegalArgumentException
                        | InvocationTargetException | NoSuchMethodException | SecurityException e) {
                    log.info("{} {}", name, e.toString());
                }
            });

            refl.getTypesAnnotatedWith(MasterRegister.class).forEach((pClass)->{
                try {
                    MasterExecute execute =  (MasterExecute)pClass.getDeclaredConstructor().newInstance();
                    MasterFactory.registerMasterLoop(execute);
                    
                } catch (InstantiationException | IllegalAccessException | IllegalArgumentException
                        | InvocationTargetException | NoSuchMethodException | SecurityException e) {
                            log.info("{}", e.toString());

                }
            });;
            
            // 启动集群
            node = cluster.start();

            rpcClient = new BoltRaftRpcFactory().createRpcClient(); // 创建rpc客户端. 
            rpcClient.init(new CliOptions()); // 初始化
        }

        public boolean isLeader() {
            return this.fsm.isLeader();
        }

        public void readIndexState(GenericClosure<State> closure) {
            getNode().readIndex(BytesUtil.EMPTY_BYTES, new ReadIndexClosure() {
                @Override
                public void run(Status status, long index, byte[] reqCtx) {
                    log.debug("readIndexState({}) {}", getServerId(), status);
                    if (status.isOk()) {
                        
                        closure.success(ss.fsm.getState());
                        closure.setValue(ss.fsm.getState());
                    } else {
                        
                        // 提交同步
                    }

                    // 回调失败
                    closure.run(status);
                }
            });
        }

        public void applyOperate(Operate op, GenericClosure<Operate> closure) {
            // 所有的提交都必须再leader进行
            if (!ss.isLeader()) {
                ss.handlerNotLeaderError(closure);
                return;
            }

            try {
                closure.setValue(op);
                final Task task = new Task();
                task.setData(
                        ByteBuffer.wrap(SerializerManager.getSerializer(SerializerManager.Hessian2).serialize(op)));
                task.setDone(closure); // 确认所有数据 一致, 不需要加锁
                StateFactory.getStateServer().getNode().apply(task);
            } catch (CodecException e) {
                String errorMsg = "Fail to encode TaskState";
                log.debug(errorMsg, e);
                closure.failure(errorMsg, PeerId.emptyPeer());
                closure.run(new Status(RaftError.EINTERNAL, errorMsg));
            }
        }

        public <T> RaftResponse<T> redirect() {
            final RaftResponse<T> response = new RaftResponse<T>();
            response.setSuccess(false);
            if (this.node != null) {
                final PeerId leader = this.node.getLeaderId();
                if (leader != null) {
                    response.setRedirect(leader);
                }
            }
            return response;
        }

        public <T> void handlerNotLeaderError(final GenericClosure<T> closure) {
            closure.failure("Not leader.", redirect().getRedirect());
            closure.run(new Status(RaftError.EPERM, "Not leader"));
        }

        private Executor createReadIndexExecutor() {
            return ThreadPoolUtil.newBuilder() //
                    .poolName("ReadIndexPool") //
                    .enableMetric(true) //
                    .coreThreads(4) //
                    .maximumThreads(4) //
                    .keepAliveSeconds(60L) //
                    .workQueue(new SynchronousQueue<>()) //
                    .threadFactory(new NamedThreadFactory("ReadIndexService", true)) //
                    .rejectedHandler(new ThreadPoolExecutor.AbortPolicy()) //
                    .build();
        }
    }

    public static void main(String[] args) throws InterruptedException, RemotingException {

        ReflectionUtils.get(ReflectionUtils.SuperClass.of(State.class));

        var rpcClient = new BoltRaftRpcFactory().createRpcClient();
        rpcClient.init(new CliOptions());
        var resp = rpcClient.invokeSync(new Endpoint("localhost", 4441), new OperateRequest(), 5000);
        log.info("{}", resp);
    }
}
